package de.informatics4kids.hopfield;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Arrays;
import java.util.Random;
import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JApplet;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JSeparator;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * Das Applet, welches die grafische Oberflche und die Zugriffe auf das
 * Neuronale Netz verwaltet.
 * <p>
 * <b>Notiz:</b> Diese Klasse ist fr Hopfield-Netze nicht direkt wichtig und
 * muss nicht vollstndig verstanden werden. Die essenziellen Teile stehen in den
 * 8 letzten Methoden ganz unten, der Rest dient nur einer mglichst intuitiven
 * Darstellung des Applets :)
 */
public class HopfieldApplet extends JApplet {

	public static final Color CELL_SET = Color.BLACK;
	public static final Color CELL_NOT_SET = Color.WHITE;

	private static final int HOPFIELD_SIZE = 8;

	private Canvas editor;
	private static final int EDITOR_SIZE = 400;
	private MathVector editorImage = new MathVector(HOPFIELD_SIZE * HOPFIELD_SIZE);

	private DefaultListModel<MathVector> imageList;

	private Hopfield hopfield;
	private JList<MathVector> listImageList;

	/**
	 * Dieser Konstruktor initialisiert das Netz.
	 */
	public HopfieldApplet() {
		hopfield = new Hopfield(HOPFIELD_SIZE * HOPFIELD_SIZE);

		JPanel contentPane = new JPanel();
		contentPane.setLayout(null);

		JLabel lblTitle = new JLabel("Hopfield");
		lblTitle.setBounds(10, 11, 46, 14);
		contentPane.add(lblTitle);

		JLabel lblEditorTitle = new JLabel("Editor");
		lblEditorTitle.setBounds(20, 36, 400, 14);
		contentPane.add(lblEditorTitle);

		editor = new Canvas();
		editor.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				int cellSize = EDITOR_SIZE / HOPFIELD_SIZE;
				int x = e.getX();
				int y = e.getY();
				int cellX = x - x % cellSize;
				int cellY = y - y % cellSize;
				int imageX = cellX / cellSize;
				int imageY = cellY / cellSize;
				int n = imageX * HOPFIELD_SIZE + imageY;
				toggleEditorCell(n);
			}
		});
		editor.setBackground(Color.WHITE);
		editor.setBounds(20, 56, EDITOR_SIZE, EDITOR_SIZE);
		contentPane.add(editor);

		JLabel lblEditorActionsTitle = new JLabel("Editor:");
		lblEditorActionsTitle.setBounds(30, 462, 40, 14);
		contentPane.add(lblEditorActionsTitle);

		JButton btnReset = new JButton("Zur\u00FCcksetzen");
		btnReset.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				resetEditorValues();
			}
		});
		btnReset.setBounds(80, 458, 116, 23);
		contentPane.add(btnReset);

		JButton btnInvert = new JButton("Invertieren");
		btnInvert.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				toggleEditorValues();
			}
		});
		btnInvert.setBounds(206, 458, 115, 23);
		contentPane.add(btnInvert);

		JButton btnRandom = new JButton("Zufall");
		btnRandom.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				randomEditorValues();
			}
		});
		btnRandom.setBounds(331, 458, 89, 23);
		contentPane.add(btnRandom);

		JLabel lblImageListTitle = new JLabel("Gelernte Bilder");
		lblImageListTitle.setBounds(482, 36, 168, 14);
		contentPane.add(lblImageListTitle);

		imageList = new DefaultListModel<MathVector>();
		listImageList = new JList<MathVector>(imageList);
		listImageList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		listImageList.setCellRenderer(new DefaultListCellRenderer() {
			@Override
			public Component getListCellRendererComponent(JList<?> list, Object value,
					int index, boolean isSelected, boolean cellHasFocus) {
				Icon icon = null;
				if (value instanceof MathVector) {
					MathVector v = (MathVector) value;
					Image img = ConversionUtils.vectorToImage(v);
					Image scaled = img.getScaledInstance(100, 100, Image.SCALE_FAST);
					icon = new ImageIcon(scaled);
				}
				JLabel label = (JLabel) super.getListCellRendererComponent(list, value,
						index, isSelected, cellHasFocus);
				label.setIcon(icon);
				label.setText("Bild #" + (index + 1));
				return label;
			}
		});
		listImageList.addListSelectionListener(new ListSelectionListener() {
			@Override
			public void valueChanged(ListSelectionEvent e) {
				System.out.println(e);
			}
		});
		listImageList.setBounds(482, 56, 168, 400);
		contentPane.add(listImageList);

		JButton btnAdd = new JButton(">");
		btnAdd.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				learnImage();
			}
		});
		btnAdd.setBounds(426, 211, 46, 23);
		contentPane.add(btnAdd);

		JButton btnEdit = new JButton("<");
		btnEdit.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				editSelectedImage();
			}
		});
		btnEdit.setBounds(426, 245, 46, 23);
		contentPane.add(btnEdit);

		JButton btnDelete = new JButton("X");
		btnDelete.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				forgetImage();
			}
		});
		btnDelete.setBounds(426, 279, 46, 23);
		contentPane.add(btnDelete);

		JSeparator separator = new JSeparator();
		separator.setBounds(20, 492, 630, 2);
		contentPane.add(separator);

		JLabel lblHopfieldActionsTitle = new JLabel("Hopfield berechnen:");
		lblHopfieldActionsTitle.setBounds(20, 505, 113, 34);
		contentPane.add(lblHopfieldActionsTitle);

		JButton btnNextItr = new JButton("N\u00E4chste Iteration");
		btnNextItr.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				performNextIteration();
			}
		});
		btnNextItr.setBounds(143, 505, 132, 34);
		contentPane.add(btnNextItr);

		JButton btnAllItrs = new JButton("Alle Iterationen");
		btnAllItrs.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				performAllIterations();
			}
		});
		btnAllItrs.setBounds(285, 505, 125, 34);
		contentPane.add(btnAllItrs);

		JButton btnSample = new JButton("Bsp");
		btnSample.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				setSample();
			}
		});
		btnSample.setBounds(490, 516, 61, 23);
		contentPane.add(btnSample);

		JButton btnResetAll = new JButton("Alles neu");
		btnResetAll.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				resetAll();
			}
		});
		btnResetAll.setBounds(561, 516, 89, 23);
		contentPane.add(btnResetAll);

		setContentPane(contentPane);
	}

	private void selectSolution(MathVector current) {
		int solution = imageList.indexOf(current);
		if (solution != -1)
			System.out.println("Lsung gefunden");
		else if ((solution = imageList.indexOf(invert(current))) != -1)
			System.out.println("Invertiere Lsung gefunden");
		else {
			System.out.println("(Noch) keine Lsung");
			listImageList.clearSelection();
		}
		listImageList.setSelectedIndex(solution);
		listImageList.ensureIndexIsVisible(solution);
	}

	private void learnImage() {
		MathVector v = getEditorValues();
		if (!imageList.contains(v)) {
			hopfield.learn(v);
			MathVector copy = new MathVector(v.dimension());
			copy.tuple = Arrays.copyOf(v.tuple, v.dimension());
			imageList.addElement(copy);
		}
	}

	private void forgetImage() {
		int selectedIndex = listImageList.getSelectedIndex();
		if (selectedIndex >= 0) {
			hopfield.forget(imageList.getElementAt(selectedIndex));
			imageList.remove(selectedIndex);
		}
	}

	private void editSelectedImage() {
		int selectedIndex = listImageList.getSelectedIndex();
		if (selectedIndex >= 0)
			setEditorValues(imageList.get(selectedIndex));
	}

	private void performNextIteration() {
		MathVector v = getEditorValues();
		MathVector next = hopfield.recognize(v);
		setEditorValues(next);
		selectSolution(next);
	}

	private void performAllIterations() {
		MathVector v = getEditorValues();
		MathVector next = hopfield.recognizeCompletely(v);
		setEditorValues(next);
		selectSolution(next);
	}

	private void setEditorCell(int n, boolean isSet) {
		if (isSet != isEditorCell(n)) {
			int val;
			if (isSet)
				val = 1;
			else
				val = -1;
			editorImage.tuple[n] = val;
			drawCell(n);
		}
	}

	private boolean isEditorCell(int n) {
		return editorImage.tuple[n] == 1;
	}

	private void toggleEditorCell(int n) {
		setEditorCell(n, !isEditorCell(n));
	}

	private void resetEditorValues() {
		for (int i = 0; i < editorImage.dimension(); i++)
			setEditorCell(i, false);
	}

	private void toggleEditorValues() {
		for (int i = 0; i < editorImage.dimension(); i++)
			toggleEditorCell(i);
	}

	private void randomEditorValues() {
		Random rnd = new Random();
		for (int i = 0; i < editorImage.dimension(); i++)
			setEditorCell(i, rnd.nextBoolean());
	}

	private MathVector getEditorValues() {
		return editorImage;
	}

	private void setEditorValues(MathVector next) {
		for (int i = 0; i < editorImage.dimension(); i++)
			setEditorCell(i, next.tuple[i] == 1);
	}

	private void drawCell(int n) {
		int x = n / HOPFIELD_SIZE;
		int y = n % HOPFIELD_SIZE;
		int cellSize = EDITOR_SIZE / HOPFIELD_SIZE;
		Graphics gr = editor.getGraphics();
		Color color;
		if (editorImage.tuple[n] == 1)
			color = CELL_SET;
		else
			color = CELL_NOT_SET;
		gr.setColor(color);
		gr.fillRect(x * cellSize, y * cellSize, cellSize, cellSize);
	}

	private MathVector invert(MathVector current) {
		for (int i = 0; i < current.dimension(); i++)
			current.tuple[i] = -current.tuple[i];
		return current;
	}

	private void setSample() {
		resetAll();
		// TODO Beispiele setzen
	}

	private void resetAll() {
		resetEditorValues();
		imageList.clear();
		hopfield.concussion();
	}
}